from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.graph import StatisticalGraph

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_GRAPH(state: State, category: str) -> None:    
    state.require(not state.in_graph, 'Already inside a graph.', 'First end current graph using EGRAPH.')
    state.require(not state.in_agent, 'Cannot define graphs inside agents.', 'First end current message using EAGENT.')
    state.require(
        not state.in_message, 
        'Cannot define graphs inside messages.', 
        'First end current message using EMESSAGE.'
    )
    state.require(not state.graph_exists(), 'Graph already exists in the current environment.')

    match category:
        case 'statistical':
            state.add_graph(StatisticalGraph())

        case _:
            state.panic(f'Incorrect operation: GRAPH {category}')
    
    state.in_graph = True


def op_EGRAPH(state: State) -> None:    
    state.require(state.in_graph, 'Not inside any graph.', 'Try defining new graphs using GRAPH.')
    if isinstance(state.last_graph, StatisticalGraph) and state.last_graph.is_agent_percent_amount_used():
        state.require(
            state.last_graph.is_size_defined(), 
            'Graph size is not defined.', 
            'Graph size must be defined to use agent percent amount.'
        )
    
    state.in_graph = False



from __future__ import annotations

from argparse import ArgumentParser
from datetime import datetime
from typing import TYPE_CHECKING, List, Tuple

from aasm.generating.python_spade import get_spade_code
from aasm.utils.exception import PanicException

if TYPE_CHECKING:
    from aasm.generating.code import Code


def get_args() -> Tuple[str, str, bool]:
    parser = ArgumentParser()
    parser.add_argument('input_path', type=str, help='path to the input file')
    parser.add_argument('-d', '--debug', action='store_true', help='toggle the compiler debug mode')
    parser.add_argument('-o', '--output-path', type=str, default='out', help='the output file path')
    args = parser.parse_args()
    return args.input_path, args.output_path, args.debug


def get_input(input_path: str) -> List[str]:
    with open(input_path, 'r') as f:
        lines = f.readlines()
    return lines


def save_output(output_path: str, code: Code) -> None:
    with open(output_path, 'w') as f:
        for code_line in code:
            f.write(code_line)


def main(input_path: str, output_path: str, debug: bool) -> None:
    lines = get_input(input_path)
    start_time = datetime.now()
    try:
        spade_code = get_spade_code(lines, indent_size=4, debug=debug)
    except PanicException as e:
        e.print()
        exit(1)
    time_delta = (datetime.now() - start_time).total_seconds()
    save_output(output_path, spade_code)
    print(f'({time_delta}s) Your results are saved in the file "{output_path}" 😎')


if __name__ == '__main__':
    input_path, output_path, debug = get_args()
    main(input_path, output_path, debug)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.intermediate.argument import Argument


class Instruction:
    def print(self) -> None:
        raise NotImplementedError()


class Comparaison(Instruction):
    def __init__(self, left: Argument, right: Argument):
        self.left = left
        self.right = right

    def print(self) -> None:
        print('Comparaison')
        self.left.print()
        self.right.print()


class IfGreaterThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)
 
    def print(self) -> None:
        print('IfGreaterThan')
        super().print()


class IfGreaterThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)
 
    def print(self) -> None:
        print('IfGreaterThanOrEqual')
        super().print()


class IfLessThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('IfLessThan')
        super().print()


class IfLessThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('IfLessThanOrEqual')
        super().print()


class IfEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('IfEqual')
        super().print()


class IfNotEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('IfNotEqual')
        super().print()


class WhileGreaterThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileGreaterThan')
        super().print()


class WhileGreaterThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileGreaterThanOrEqual')
        super().print()


class WhileLessThan(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileLessThan')
        super().print()


class WhileLessThanOrEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileLessThanOrEqual')
        super().print()


class WhileEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileEqual')
        super().print()


class WhileNotEqual(Comparaison):
    def __init__(self, left: Argument, right: Argument):
        super().__init__(left=left, right=right)

    def print(self) -> None:
        print('WhileNotEqual')
        super().print()


class MathOperation(Instruction):
    def __init__(self, dst: Argument, num: Argument):
        self.dst = dst
        self.num = num

    def print(self) -> None:
        print('MathOperation')
        self.dst.print()
        self.num.print()


class Multiply(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print('Multiply')
        super().print()


class Divide(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print('Divide')
        super().print()


class Add(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print('Add')
        super().print()


class Subtract(MathOperation):
    def __init__(self, dst: Argument, num: Argument):
        super().__init__(dst=dst, num=num)

    def print(self) -> None:
        print('Subtract')
        super().print()


class ListElementAccess(Instruction):
    def __init__(self, list_: Argument, element: Argument):
        self.list_ = list_
        self.element = element

    def print(self) -> None:
        print('ListElementAccess')
        self.list_.print()
        self.element.print()


class AddElement(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print('AddElement')
        super().print()


class RemoveElement(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print('RemoveElement')
        super().print()


class IfInList(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print('IfInList')
        super().print()


class IfNotInList(ListElementAccess):
    def __init__(self, list_: Argument, element: Argument):
        super().__init__(list_=list_, element=element)

    def print(self) -> None:
        print('IfNotInList')
        super().print()


class RemoveNElements(Instruction):
    def __init__(self, list_: Argument, num: Argument):
        self.list_ = list_
        self.num = num

    def print(self) -> None:
        print('RemoveNElements')
        self.list_.print()
        self.num.print()


class Subset(Instruction):
    def __init__(self, dst_list: Argument, src_list: Argument, num: Argument):
        self.dst_list = dst_list
        self.src_list = src_list
        self.num = num

    def print(self) -> None:
        print('Subset')
        self.dst_list.print()
        self.src_list.print()
        self.num.print()


class Set(Instruction):
    def __init__(self, dst: Argument, value: Argument):
        self.dst = dst
        self.value = value

    def print(self) -> None:
        print('Set')
        self.dst.print()
        self.value.print()


class Clear(Instruction):
    def __init__(self, list_: Argument):
        self.list_ = list_

    def print(self) -> None:
        print('Clear')
        self.list_.print()


class Length(Instruction):
    def __init__(self, dst: Argument, list_: Argument):
        self.dst = dst
        self.list_ = list_

    def print(self) -> None:
        print('Length')
        self.dst.print()
        self.list_.print()


class Send(Instruction):
    def __init__(self, receivers: Argument):
        self.receivers = receivers

    def print(self) -> None:
        print('Send')
        self.receivers.print()


class Round(Instruction):
    def __init__(self, dst: Argument):
        self.dst = dst

    def print(self) -> None:
        print('Round')
        self.dst.print()


class UniformDist(Instruction):
    def __init__(self, dst: Argument, a: Argument, b: Argument):
        self.dst = dst
        self.a = a
        self.b = b

    def print(self) -> None:
        print('UniformDist')
        self.dst.print()
        self.a.print()
        self.b.print()


class NormalDist(Instruction):
    def __init__(self, dst: Argument, mean: Argument, std_dev: Argument):
        self.dst = dst
        self.mean = mean
        self.std_dev = std_dev

    def print(self) -> None:
        print('NormalDist')
        self.dst.print()
        self.mean.print()
        self.std_dev.print()


class ExpDist(Instruction):
    def __init__(self, dst: Argument, lambda_: Argument):
        self.dst = dst
        self.lambda_ = lambda_

    def print(self) -> None:
        print('ExpDist')
        self.dst.print()
        self.lambda_.print()


class ListRead(Instruction):
    def __init__(self, dst: Argument, list_: Argument, idx: Argument):
        self.dst = dst
        self.list_ = list_
        self.idx = idx


class ListWrite(Instruction):
    def __init__(self, list_: Argument, idx: Argument, value: Argument):
        self.list_ = list_
        self.idx = idx
        self.value = value



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List

from aasm.utils.iteration import zip_consecutive_pairs

if TYPE_CHECKING:
    from aasm.intermediate.behaviour import (Behaviour, CyclicBehaviour,
                                             MessageReceivedBehaviour,
                                             OneTimeBehaviour, SetupBehaviour)


class InitFloatParam:
    def __init__(self, name: str, value: str):
        self.name: str = name
        self.value: str = value
        
    def print(self) -> None:
        print(f'InitFloatParam {self.name} = {self.value}')


class DistNormalFloatParam:
    def __init__(self, name: str, mean: str, std_dev: str):
        self.name: str = name
        self.mean: str = mean
        self.std_dev: str = std_dev

    def print(self) -> None:
        print(f'DistNormalFloatParam {self.name} = normal(mean={self.mean}, std_dev={self.std_dev})')


class DistExpFloatParam:
    def __init__(self, name: str, lambda_: str):
        self.name: str = name
        self.lambda_: str = lambda_

    def print(self) -> None:
        print(f'DistExpFloatParam {self.name} = exp(lambda={self.lambda_})')


class DistUniformFloatParam:
    def __init__(self, name: str, a: str, b: str):
        self.name: str = name
        self.a: str = a
        self.b: str = b

    def print(self) -> None:
        print(f'DistUniformFloatParam {self.name} = uniform(a={self.a}, b={self.b})')


class EnumParam:
    def __init__(self, name: str, enums: List[str]):
        self.name: str = name
        self.enum_values: List[EnumValue] = [
            EnumValue(name, value, percentage) for value, percentage in zip_consecutive_pairs(enums)
        ]
        
    def print(self) -> None:
        print(f'EnumParam {self.name} = {self.enum_values}')


class EnumValue:
    def __init__(self, from_enum: str, value: str, percentage: str):
        self.from_enum: str = from_enum
        self.value: str = value
        self.percentage: str = percentage
        
    def __str__(self) -> str:
        return f'({self.value}, {self.percentage}; from_enum={self.from_enum})'


class MessageListParam:
    def __init__(self, name: str):
        self.name: str = name
        
    def print(self) -> None:
        print(f'MessageListParam {self.name} = []')


class ConnectionListParam:
    def __init__(self, name: str):
        self.name: str = name
        
    def print(self) -> None:
        print(f'ConnectionListParam {self.name} = []')


class FloatListParam:
    def __init__(self, name: str):
        self.name: str = name
    
    def print(self) -> None:
        print(f'FloatListParam {self.name} = []')


class Agent:
    RESERVED_CONNECTION_LIST_PARAMS = [ 'connections' ]
    RESERVED_FLOAT_PARAMS = [ 'connCount', 'msgRCount', 'msgSCount' ]
    
    def __init__(self, name: str):
        self.name: str = name
        self.init_floats: Dict[str, InitFloatParam] = {}
        self.dist_normal_floats: Dict[str, DistNormalFloatParam] = {}
        self.dist_exp_floats: Dict[str, DistExpFloatParam] = {}
        self.dist_unifrom_floats: Dict[str, DistUniformFloatParam] = {}
        self.enums: Dict[str, EnumParam] = {}
        self.connection_lists: Dict[str, ConnectionListParam] = {}
        self.message_lists: Dict[str, MessageListParam] = {}
        self.float_lists: Dict[str, FloatListParam] = {}
        self.setup_behaviours: Dict[str, SetupBehaviour] = {}
        self.one_time_behaviours: Dict[str, OneTimeBehaviour] = {}
        self.cyclic_behaviours: Dict[str, CyclicBehaviour] = {}
        self.message_received_behaviours: Dict[str, MessageReceivedBehaviour] = {}
        self._last_modified_behaviour: Behaviour | None = None
    
    @property
    def last_behaviour(self) -> Behaviour:
        if self._last_modified_behaviour is None:
            raise Exception('No behaviour added to agent')
        return self._last_modified_behaviour
    
    @property
    def param_names(self) -> List[str]:
        return [ *Agent.RESERVED_CONNECTION_LIST_PARAMS,
                 *Agent.RESERVED_FLOAT_PARAMS,
                 *list(self.init_floats), 
                 *list(self.dist_normal_floats),
                 *list(self.dist_exp_floats),
                 *list(self.dist_unifrom_floats),
                 *list(self.enums), 
                 *list(self.connection_lists),
                 *list(self.message_lists),
                 *list(self.float_lists) ]
    
    @property
    def behaviour_names(self) -> List[str]:
        return [ *list(self.setup_behaviours),
                 *list(self.one_time_behaviours),
                 *list(self.cyclic_behaviours),
                 *list(self.message_received_behaviours) ]
    
    @property
    def float_param_names(self) -> List[str]:
        return [ *list(self.init_floats),
                 *list(self.dist_normal_floats), 
                 *list(self.dist_exp_floats),
                 *list(self.dist_unifrom_floats) ]
    
    def add_init_float(self, float_param: InitFloatParam) -> None:
        self.init_floats[float_param.name] = float_param

    def add_dist_normal_float(self, float_param: DistNormalFloatParam) -> None:
        self.dist_normal_floats[float_param.name] = float_param
    
    def add_dist_exp_float(self, float_param: DistExpFloatParam) -> None:
        self.dist_exp_floats[float_param.name] = float_param
        
    def add_dist_uniform_float(self, float_param: DistUniformFloatParam) -> None:
        self.dist_unifrom_floats[float_param.name] = float_param
    
    def add_enum(self, enum_param: EnumParam) -> None:
        self.enums[enum_param.name] = enum_param
    
    def add_connection_list(self, list_param: ConnectionListParam) -> None:
        self.connection_lists[list_param.name] = list_param
    
    def add_message_list(self, list_param: MessageListParam) -> None:
        self.message_lists[list_param.name] = list_param
        
    def add_float_list(self, list_param: FloatListParam) -> None:
        self.float_lists[list_param.name] = list_param
    
    def add_setup_behaviour(self, behaviour: SetupBehaviour) -> None:
        self.setup_behaviours[behaviour.name] = behaviour
        self._last_modified_behaviour = behaviour
    
    def add_one_time_behaviour(self, behaviour: OneTimeBehaviour) -> None:
        self.one_time_behaviours[behaviour.name] = behaviour
        self._last_modified_behaviour = behaviour
    
    def add_cyclic_behaviour(self, behaviour: CyclicBehaviour) -> None:
        self.cyclic_behaviours[behaviour.name] = behaviour
        self._last_modified_behaviour = behaviour
    
    def add_message_received_behaviour(self, behaviour: MessageReceivedBehaviour) -> None:
        self.message_received_behaviours[behaviour.name] = behaviour
        self._last_modified_behaviour = behaviour

    def param_exists(self, name: str) -> bool:
        return name in self.param_names
    
    def behaviour_exists(self, name: str) -> bool:
        return name in self.behaviour_names

    def name_exists(self, name: str) -> bool:
        return self.param_exists(name) or self.behaviour_exists(name)
        
    def behaviour_for_template_exists(self, msg_type: str, msg_performative: str) -> bool:
        for msg_rcv_behav in self.message_received_behaviours.values():
            if msg_rcv_behav.received_message.type == msg_type and msg_rcv_behav.received_message.performative == msg_performative:
                return True
        return False
    
    def print(self) -> None:
        print(f'Agent {self.name}')
        for init_float_param in self.init_floats.values():
            init_float_param.print()
        for dist_normal_float_param in self.dist_normal_floats.values():
            dist_normal_float_param.print()
        for dist_exp_float_param in self.dist_exp_floats.values():
            dist_exp_float_param.print()
        for dist_uniform_float_param in self.dist_unifrom_floats.values():
            dist_uniform_float_param.print()
        for enum_param in self.enums.values():
            enum_param.print()
        for connection_list_param in self.connection_lists.values():
            connection_list_param.print()
        for message_list_param in self.message_lists.values():
            message_list_param.print()
        for float_list_param in self.float_lists.values():
            float_list_param.print()
        for setup_behaviour in self.setup_behaviours.values():
            setup_behaviour.print()
        for one_time_behaviour in self.one_time_behaviours.values():
            one_time_behaviour.print()
        for cyclic_behaviour in self.cyclic_behaviours.values():
            cyclic_behaviour.print()
        for message_received_behaviour in self.message_received_behaviours.values():
            message_received_behaviour.print()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Add, Divide, Multiply, Subtract

if TYPE_CHECKING:
    from aasm.parsing.state import State


def handle_math_statement(state: State, op: str, arg1: str, arg2: str) -> None:
    state.require(state.in_action, 'Not inside any action', f'{op} can be used inside actions.')
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.math_context(rhs),
        'Mismatched types in the math statement.', 
        f'ARG1 {lhs.explain()}, ARG2 {rhs.explain()}'
    )

    match op:
        case 'ADD':
            state.last_action.add_instruction(Add(lhs, rhs))

        case 'SUBT':
            state.last_action.add_instruction(Subtract(lhs, rhs))

        case 'MULT':
            state.last_action.add_instruction(Multiply(lhs, rhs))

        case 'DIV':
            state.last_action.add_instruction(Divide(lhs, rhs))

        case _:
            state.panic(f'Unexpected error: {op} {arg1} {arg2}')



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_EBLOCK(state: State) -> None:            
    state.require(state.in_action, 'Cannot end blocks outside actions.')
    state.require(state.last_action.nested_blocks_count > 0, 'No more blocks to close', 'Try removing this statement.')
    
    state.last_action.end_block()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.message import Message
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_MESSAGE(state: State, msg_type: str, msg_performative: str) -> None:
    state.require(not state.in_message, 'Already inside a message.', 'First end current message using EMESSAGE.')
    state.require(not state.in_agent, 'Cannot define messages inside agents.', 'First end current agent using EAGENT.')
    state.require(not state.in_graph, 'Cannot define messages inside graphs.', 'First end current graph using EGRAPH.')
    state.require(
        not state.message_exists(msg_type, msg_performative), 
        f'Message {msg_type}/{msg_performative} already exists in the current environment.'
    )
    state.require(
        is_valid_name(msg_type), 
        f'{msg_type} is not a correct name.', 
        f'Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.'
    )
    state.require(
        is_valid_name(msg_performative), 
        f'{msg_performative} is not a correct name.', 
        f'Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.'
    )
    
    state.in_message = True
    state.add_message(Message(msg_type, msg_performative))


def op_EMESSAGE(state: State) -> None:    
    state.require(state.in_message, 'Not inside any message.', 'Try defining new messages using MESSAGE.')
    
    state.in_message = False



from __future__ import annotations

from copy import deepcopy
from pprint import pprint
from typing import TYPE_CHECKING, Dict, Generator, List, NoReturn, Tuple

from aasm.utils.exception import PanicException

if TYPE_CHECKING:
    from aasm.intermediate.action import Action
    from aasm.intermediate.agent import Agent
    from aasm.intermediate.behaviour import Behaviour
    from aasm.intermediate.graph import Graph
    from aasm.intermediate.message import Message


class ParsedData:
    def __init__(self, agents: List[Agent], messages: List[Message], graph: Graph | None):
        self.agents: List[Agent] = agents
        self.messages: List[Message] = messages
        self.graph: Graph | None = graph


class State:
    def __init__(self, lines: List[str], debug: bool):
        self.debug: bool = debug
        self.lines: List[str] = lines
        self.line_num: int = 0
        self.in_agent: bool = False
        self.in_message: bool = False
        self.in_behaviour: bool = False
        self.in_action: bool = False
        self.in_graph: bool = False
        self.agents: Dict[str, Agent] = {}
        self.messages: Dict[Tuple[str, str], Message] = {}
        self.graph: Graph | None = None
        
    @property
    def last_agent(self) -> Agent:
        return self.agents[list(self.agents.keys())[-1]]
    
    @property
    def last_behaviour(self) -> Behaviour:
        return self.last_agent.last_behaviour
    
    @property
    def last_action(self) -> Action:
        return self.last_agent.last_behaviour.last_action
    
    @property
    def last_message(self) -> Message:
        return self.messages[list(self.messages.keys())[-1]]
    
    @property
    def last_graph(self) -> Graph:
        if self.graph is None:
            raise Exception('Graph is not defined')
        return self.graph
    
    def add_agent(self, agent: Agent) -> None:
        self.agents[agent.name] = agent
        
    def add_message(self, message: Message) -> None:
        self.messages[(message.type, message.performative)] = message

    def add_graph(self, graph: Graph) -> None:
        self.graph = graph
        
    def agent_exists(self, name: str) -> bool:
        return name in self.agents
    
    def message_exists(self, msg_type: str, msg_performative: str) -> bool:
        return (msg_type, msg_performative) in self.messages

    def graph_exists(self) -> bool:
        return self.graph is not None
    
    def get_message_instance(self, msg_type: str, msg_performative: str) -> Message:
        return deepcopy(self.messages[(msg_type, msg_performative)])
            
    def tokens_from_lines(self) -> Generator[list[str], None, None]:
        for line in self.lines:
            self.line_num += 1
            uncommented = line.split('#')[0]
            tokens = [token.strip() for token in uncommented.replace(',', ' ').split()]
            if tokens:
                tokens[0] = tokens[0].upper()
                yield tokens
                
    def print(self) -> None:
        if self.messages:
            print('- Messages:')
            for message in self.messages.values():
                message.print()
                print('')
        if self.agents:
            print('- Agents:')
            for agent in self.agents.values():
                agent.print()
                print('')
        if self.graph:
            print('- Graph:')
            self.graph.print()
    
    def verify_end_state(self) -> None:
        if self.in_agent:
            self.panic('Missing EAGENT')
        elif self.in_message:
            self.panic('Missing EMESSAGE')
        elif self.in_graph:
            self.panic('Missing EGRAPH')
            
    def get_parsed_data(self) -> ParsedData:
        self.verify_end_state()
        if self.debug:
            pprint(self.__dict__)
            self.print()
        return ParsedData(list(self.agents.values()), list(self.messages.values()), self.graph)

    def panic(self, reason: str, suggestion: str = '') -> NoReturn:
        if self.debug:
            pprint(self.__dict__)
            self.print()
        place = f'Error in line {self.line_num}: {self.lines[self.line_num - 1].strip()}'
        raise PanicException(place, reason, suggestion)

    def require(self, expr: bool, msg_on_error: str, suggestion_on_error: str = '') -> None:
        if not expr:
            self.panic(msg_on_error, suggestion_on_error)



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.declaration import Declaration
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_DECL(state: State, name: str, value: str) -> None:            
    state.require(state.in_action, 'Cannot declare variables outside actions.')
    state.require(
        is_valid_name(name), 
        f'{name} is not a correct name.', 
        f'Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.'
    )
    state.require(not state.last_agent.param_exists(name), f'{name} is already defined in current agent.')
    state.require(
        not state.last_action.is_declaration_in_scope(name), 
        f'{name} is already declared in current action scope.'
    )
    lhs = Argument(state, name)
    rhs = Argument(state, value)
    state.require(
        lhs.declaration_context(rhs), 
        'Mismatched types in the declaration context.', 
        f'ARG1 {lhs.explain()}, ARG2 {rhs.explain()}'
    )
    
    state.last_action.add_declaration(Declaration(lhs, rhs))
